#include "DoubleVoxmap.h"
#include <math.h>

static void check_(DoubleVoxmap *this, ulong i, ulong j, ulong k);

static ulong add_subtract_(
    DoubleVoxmap **p_this,
    DoubleVoxmap *a,
    DoubleVoxmap *b,
    bool subtract);


void double_voxmap_new(
    DoubleVoxmap **p_this,
    ulong width,
    ulong height,
    ulong depth) {
  DoubleVoxmap *this;

  if (width == 0) {
    throw("Width must be nonzero");
  }

  if (height== 0) {
    throw("Height must be nonzero");
  }

  if (depth== 0) {
    throw("Depth must be nonzero");
  }

  this = newo(p_this);

  this->width = width;
  this->height = height;
  this->depth = depth;
  this->numVoxels = width * height * depth;

  newt3(&this->_voxels, width, height, depth);
}

void double_voxmap_new_add(
    DoubleVoxmap **p_this,
    DoubleVoxmap *a,
    DoubleVoxmap *b) {
  add_subtract_(p_this, a, b, false);
}

void double_voxmap_new_subtract(
    DoubleVoxmap **p_this,
    DoubleVoxmap *a,
    DoubleVoxmap *b) {
  add_subtract_(p_this, a, b, true);
}

void double_voxmap_delete(DoubleVoxmap **p_this) {
  DoubleVoxmap *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletet3v(&this->_voxels, this->width, this->height);

  deleteo(p_this);
}

void double_voxmap_print(DoubleVoxmap *this, Printer *p) {
  ulong ii;
  ulong j;
  ulong k;

  sprint(p, "DoubleVoxmap");
  rprint(p, "Width: %lu", this->width);
  rprint(p, "Height: %lu", this->height);
  rprint(p, "Depth: %lu", this->depth);

  sprint(p, "Contents");

  for (k = 0; k < this->depth; k++) {
    sprint(p, "Slice %lu", k);

    for (j = 0; j < this->height; j++) {
      sprint(p, "Row %lu", j);

      for (ii = 0; ii < this->width; ii++) {
        rprint(p, "[%lu] %.16g", ii, this->_voxels[ii][j][k]);
      }

      fprint(p);
    }

    fprint(p);
  }

  fprint(p);
  fprint(p);
}

double double_voxmap_get(DoubleVoxmap *this, ulong i, ulong j, ulong k) {
  check_(this, i, j, k);
  return this->_voxels[i][j][k];
}

void double_voxmap_set(
    DoubleVoxmap *this,
    ulong i,
    ulong j,
    ulong k,
    double d) {
  check_(this, i, j, k);
  this->_voxels[i][j][k] = d;
}

double double_voxmap_max_abs_diff(DoubleVoxmap *a, DoubleVoxmap *b) {
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double absdiff;
  double max;

  width = a->width;
  height = a->height;
  depth = a->depth;

  if (b->width != width) {
    throw("Voxmaps have different widths (%lu v %lu)", width, b->width);
  }

  if (b->height != height) {
    throw("Voxmaps have different heights (%lu v %lu)", height, b->height);
  }

  if (b->depth != depth) {
    throw("Voxmaps have different depths (%lu v %lu)", depth, b->depth);
  }

  for (max = 0, i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        absdiff = fabs(a->_voxels[i][j][k] - b->_voxels[i][j][k]);
        max = maximum(max, absdiff);
      }
    }
  }

  return max;
}


static void check_(DoubleVoxmap *this, ulong i, ulong j, ulong k) {
  if (i >= this->width) {
    throw("i (%lu) must be less than width (%lu)", i, this->width);
  }

  if (j >= this->height) {
    throw("j (%lu) must be less than height (%lu)", j, this->height);
  }

  if (k >= this->depth) {
    throw("k (%lu) must be less than depth (%lu)", k, this->depth);
  }
}

static ulong add_subtract_(
    DoubleVoxmap **p_this,
    DoubleVoxmap *a,
    DoubleVoxmap *b,
    bool subtract) {
  ulong depth;
  ulong height;
  ulong i;
  ulong j;
  ulong k;
  ulong width;
  double da;
  double db;
  double r;
  DoubleVoxmap *this;

  width = a->width;
  height = a->height;
  depth = a->depth;

  if (b->width != width) {
    throw("Voxmaps have different widths (%lu v %lu)", width, b->width);
  }

  if (b->height != height) {
    throw("Voxmaps have different heights (%lu v %lu)", height, b->height);
  }

  if (b->depth != depth) {
    throw("Voxmaps have different depths (%lu v %lu)", depth, b->depth);
  }

  double_voxmap_new(p_this, width, height, depth);
  this = *p_this;

  for (i = 0; i < width; i++) {
    for (j = 0; j < height; j++) {
      for (k = 0; k < depth; k++) {
        da = a->_voxels[i][j][k];
        db = b->_voxels[i][j][k];

        if (subtract) {
          r = da - db;
        } else {
          r = da + db;
        }

        this->_voxels[i][j][k] = r;
      }
    }
  }

  return 0;
}
