/* Object class for readind and writing XML file tokens in a streaming fashion.
 */
#ifndef XML_FILE_H
#define XML_FILE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "File.h"

typedef struct {
  bool wrapped;
  bool readTagStart;
  bool isTag;
  bool isStartTag;
  bool isEndTag;
  bool isEmptyElementTag;
  char *_token;
  char *_tagName;
  char *_attributes;
  ulong maxTokenChars;
  ulong maxTagNameChars;
  ulong maxAttributesChars;
  File *_file;
} XmlFile;

/* Constructor, opening a file from filename and mode.
 */
void xml_file_new(XmlFile **p_this, char *filename, char *mode);

/* Constructor, wrapping an existing file.
 */
void xml_file_new_wrap_existing(XmlFile **p_this, File *file);

/* Constructor of a memory XmlFile, reading the contents of the current element
 * of another XmlFile, i.e. one whose start tag has already been read.
 */
void xml_file_new_memory_read_current_element(
    XmlFile **p_this,
    XmlFile *other);

/* Constructor of a memory XmlFile, reading its contents from another XmlFile
 * after confirming that the next element has the specified name.
 */
void xml_file_new_memory_read_element(
    XmlFile **p_this,
    XmlFile *other,
    char *element_name);

/* Destructor.
 */
void xml_file_delete(XmlFile **p_this);

/* Read a token, which is either a tag markup or content between tags. Data
 * about the current token is kept in the object.
 */
void xml_file_read_token(XmlFile *this);

/* Try to read a token.
 */
bool xml_file_try_read_token(XmlFile *this);

/* Parse a tag token. For performance reasons, tag tokens are not parsed unless
 * this is called.
 */
void xml_file_parse_tag(XmlFile *this);

/* Skip to the end tag of the specified tag, calling the callback function for
 * each token read along the way.
 */
void xml_file_skip_to_end_tag(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x);

/* Skip to the specified element, calling the callback function for each token
 * read along the way.
 */
void xml_file_skip_to_element(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x);

/* Try to skip to the specified element.
 */
bool xml_file_try_skip_to_element(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x);

/* Read a token and confirm that it is a tag with the given tag name.
 */
void xml_file_read_tag(XmlFile *this, char *tag_name);

/* Read a token and confirm that it is a tag with the given tag name and has
 * the given attributes string.
 */
void xml_file_read_tag_attributes(
    XmlFile *this,
    char *tag_name,
    char *attributes);

/* Read a token and confirm that it is a start tag with the given tag name.
 */
void xml_file_read_start_tag(XmlFile *this, char *tag_name);

/* Read a token and confirm that it is a start tag with the given tag name and
 * has the given attributes string.
 */
void xml_file_read_start_tag_attributes(
    XmlFile *this,
    char *tag_name,
    char *attributes);

/* Skip the next element, confirming that it has the specified name.
 */
void xml_file_skip_element(XmlFile *this, char *element_name);

/* Read the contents of the next element, confirming that it has the specified
 * name.
 */
void xml_file_read_element(
    XmlFile *this,
    char *element_name,
    char **pcontents);

/* Print information about the token that has just been read. This will parse
 * the token, if it is a tag.
 */
void xml_file_print(XmlFile *this, Printer *p);

/* Write a token. Currently this is simple a string write.
 */
void xml_file_write_token(XmlFile *this, char *fmt, ...);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
