#include "double_voxmap_kernel.h"

#define MIN_ 100000

typedef struct {
  ulong numV;
  double ***temp1;
  double ***temp2;
  DoubleVoxmap *input;
  DoubleVoxmap *output;
  Kernel *kernel;
  double (*read)(ulong in, void *x);
  void (*write)(ulong out, double v, void *x);
} Pass_;

typedef struct {
  ulong u;
  ulong v;
  Pass_ *pass;
} KernelPass_;

static void apply_(Parallel *par);

static double read_x_(ulong in, void *x);

static void write_x_(ulong out, double v, void *x);

static double read_y_(ulong in, void *x);

static void write_y_(ulong out, double v, void *x);

static double read_z_(ulong in, void *x);

static void write_z_(ulong out, double v, void *x);


DoubleVoxmap *double_voxmap_new_apply_xyz_kernels(
    DoubleVoxmap **p_this,
    DoubleVoxmap *other,
    Kernel *x_kernel,
    Kernel *y_kernel,
    Kernel *z_kernel) {
  ulong in_depth;
  ulong in_height;
  ulong in_width;
  ulong min_parallel;
  ulong out_depth;
  ulong out_height;
  ulong out_width;
  ulong temp_1_depth;
  ulong temp_1_height;
  ulong temp_1_width;
  ulong temp_2_depth;
  ulong temp_2_height;
  ulong temp_2_width;
  double ***_temp_1 = null;
  double ***_temp_2 = null;
  Pass_ pass;
  DoubleVoxmap *this;

  in_width = other->width;
  in_height = other->height;
  in_depth = other->depth;

  if (x_kernel->nIn != in_width) {
    throw("X kernel n_in (%lu) must equal input voxmap width (%lu)",
        x_kernel->nIn,
        in_width);
  }

  if (y_kernel->nIn != in_height) {
    throw("Y kernel n_in (%lu) must equal input voxmap height (%lu)",
        y_kernel->nIn,
        in_height);
  }

  if (z_kernel->nIn != in_depth) {
    throw("Z kernel n_in (%lu) must equal input voxmap depth (%lu)",
        z_kernel->nIn,
        in_depth);
  }

  out_width = x_kernel->nOut;
  out_height = y_kernel->nOut;
  out_depth = z_kernel->nOut;

  temp_1_width = out_width;
  temp_1_height = in_height;
  temp_1_depth = in_depth;

  temp_2_width = out_width;
  temp_2_height = out_height;
  temp_2_depth = in_depth;

  pass.input = other;
  double_voxmap_new(
      p_this,
      out_width,
      out_height,
      out_depth);
  pass.output = this = *p_this;

  newt3(&_temp_1, temp_1_width, temp_1_height, temp_1_depth);
  pass.temp1 = _temp_1;

  newt3(&_temp_2, temp_2_width, temp_2_height, temp_2_depth);
  pass.temp2 = _temp_2;

  pass.numV = temp_1_depth;
  pass.kernel = x_kernel;
  pass.read = read_x_;
  pass.write = write_x_;
  min_parallel = pass.kernel->minParallel / pass.numV;
  parallel_for(temp_1_height, min_parallel, apply_, &pass);

  pass.numV = temp_2_width;
  pass.kernel = y_kernel;
  pass.read = read_y_;
  pass.write = write_y_;
  min_parallel = pass.kernel->minParallel / pass.numV;
  parallel_for(temp_2_depth, min_parallel, apply_, &pass);

  pass.numV = out_height;
  pass.kernel = z_kernel;
  pass.read = read_z_;
  pass.write = write_z_;
  min_parallel = pass.kernel->minParallel / pass.numV;
  parallel_for(out_width, min_parallel, apply_, &pass);

  deletet3v(&_temp_2, temp_2_width, temp_2_height);
  deletet3v(&_temp_1, temp_1_width, temp_1_height);

  return this;
}

DoubleVoxmap *double_voxmap_new_apply_multiple_xyz_kernels(
    DoubleVoxmap **p_this,
    DoubleVoxmap *other,
    ulong num_kernel_pairs,
    Kernel **x_kernels,
    Kernel **y_kernels,
    Kernel **z_kernels) {
  byte k;
  DoubleVoxmap *_temp_1 = null;
  DoubleVoxmap *_temp_2 = null;

  for (k = 0; k < num_kernel_pairs; k++) {
    double_voxmap_new_apply_xyz_kernels(
        &_temp_2,
        k == 0 ? other : _temp_1,
        x_kernels[k],
        y_kernels[k],
        z_kernels[k]);
    double_voxmap_delete(&_temp_1);
    donate(&_temp_2, &_temp_1);
  }

  donate(&_temp_1, p_this);

  return *p_this;
}


static void apply_(Parallel *par) {
  ulong u;
  ulong v;
  KernelPass_ kpass;
  Pass_ *pass;

  pass = par->x;

  kpass.pass = pass;

  for (u = par->start; u < par->end; u++) {
    kpass.u = u;

    for (v = 0; v < pass->numV; v++) {
      kpass.v = v;
      kernel_apply(pass->kernel, pass->read, pass->write, &kpass);
    }
  }
}

static double read_x_(ulong in, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = in;
  j = kpass->u;
  k = kpass->v;

  return kpass->pass->input->_voxels[i][j][k];
}

static void write_x_(ulong out, double v, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = out;
  j = kpass->u;
  k = kpass->v;

  kpass->pass->temp1[i][j][k] = v;
}

static double read_y_(ulong in, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = kpass->v;
  j = in;
  k = kpass->u;

  return kpass->pass->temp1[i][j][k];
}

static void write_y_(ulong out, double v, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = kpass->v;
  j = out;
  k = kpass->u;

  kpass->pass->temp2[i][j][k] = v;
}

static double read_z_(ulong in, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = kpass->u;
  j = kpass->v;
  k = in;

  return kpass->pass->temp2[i][j][k];
}

static void write_z_(ulong out, double v, void *x) {
  ulong i;
  ulong j;
  ulong k;
  KernelPass_ *kpass;

  kpass = x;

  i = kpass->u;
  j = kpass->v;
  k = out;

  kpass->pass->output->_voxels[i][j][k] = v;
}
