#include "resize_pvl_nc.h"
#include <math.h>

int main(int argc, char *argv[]) {
  char *stub;
  char *_s = null;
  ulong in_depth;
  ulong in_height;
  ulong in_width;
  ulong m;
  ulong out_depth;
  ulong out_height;
  ulong out_width;
  double extra_sharp;
  double k;
  double k_recip;
  ByteVoxmap *_in = null;
  ByteVoxmap *_out = null;
  DoubleVoxmap *_din = null;
  DoubleVoxmap *_dtemp = null;
  DoubleVoxmap *_dout = null;
  Logger *_l = null;

  args_description("Resize a .pvl.nc voxmap file");
  args_string("input-filename-stub", "Input base filename (without .pvl.nc)");
  args_string(
      "output-filename-stub",
      "Resized output base filename (without .pvl.nc)");
  args_optional_integer(
      "width",
      'w',
      ARGS_INTEGER_NO_DEFAULT,
      "The width of the output voxmap, in pixels");
  args_optional_integer(
      "height",
      'H',
      ARGS_INTEGER_NO_DEFAULT,
      "The height of the output voxmap, in pixels");
  args_optional_integer(
      "depth",
      'd',
      ARGS_INTEGER_NO_DEFAULT,
      "The depth of the output voxmap, in pixels");
  args_optional_real(
      "scale-factor",
      'k',
      ARGS_REAL_NO_DEFAULT,
      "The scale factor k, where k > 1 for upsizing");
  args_optional_real(
      "recip-scale-factor",
      'r',
      ARGS_REAL_NO_DEFAULT,
      "The reciprocal 1/k of the scale factor k; useful for specifying "
      "exact integer values of 1/k");
  string_new_copy(&_s, "The resize method, which must be one of:");

  for (m = 0; m < MAGIC_NUM_RESIZERS; m++) {
    string_renew_append_f(
        &_s,
        "\n\t%s: %s",
        MAGIC_RESIZERS[2 * m],
        MAGIC_RESIZERS[2 * m + 1]);
  }

  args_optional_string("method", 'm', MAGIC_RESIZER_MAGIC_KERNEL_SHARP, _s);
  string_delete(&_s);
  args_optional_integer(
      "a",
      'a',
      3,
      "The generation 'a' of Magic Kernel Sharp, the Magic Kernel, or Lanczos "
      "kernel, as described in the paper above (valid values for each are "
      "listed under METHOD above)");
  args_optional_integer(
      "v",
      'v',
      3,
      "The version of approximation 'v' to the exact Sharp filter in "
      "Magic Kernel Sharp (valid values are listed above)");
  args_optional_real(
      "extra-sharpening",
      's',
      0,
      "Extra sharpening to be applied to the voxmap");
  args_parse(argc, argv);

  logger_new(&_l, stdout);

  stub = args_get_string("input-filename-stub");
  slog(_l, "Loading %s.pvl.nc", stub);
  byte_voxmap_new_load_pvl_nc(&_in, stub);
  flog(_l);

  slog(_l, "Converting byte voxmap to double");
  double_voxmap_new_from_byte(&_din, _in);
  byte_voxmap_delete(&_in);
  flog(_l);

  in_width = _din->width;
  in_height = _din->height;
  in_depth = _din->depth;

  if (args_set_veto("scale-factor", "recip-scale-factor")) {
    k = args_get_positive_real("scale-factor");

    if (args_set("width")) {
      out_width = args_get_positive_integer("width");
    } else {
      out_width = (ulong)ceil(k * in_width);
    }

    if (args_set("height")) {
      out_height = args_get_positive_integer("height");
    } else {
      out_height = (ulong)ceil(k * in_height);
    }

    if (args_set("depth")) {
      out_depth = args_get_positive_integer("depth");
    } else {
      out_depth = (ulong)ceil(k * in_depth);
    }
  } else if (args_set("recip-scale-factor")) {
    k = 1 / (k_recip = args_get_positive_real("recip-scale-factor"));

    if (args_set("width")) {
      out_width = args_get_positive_integer("width");
    } else {
      out_width = (ulong)ceil(in_width / k_recip);
    }

    if (args_set("height")) {
      out_height = args_get_positive_integer("height");
    } else {
      out_height = (ulong)ceil(in_height / k_recip);
    }

    if (args_set("depth")) {
      out_depth = args_get_positive_integer("depth");
    } else {
      out_depth = (ulong)ceil(in_depth / k_recip);
    }
  } else if (args_set("width")) {
    out_width = args_get_positive_integer("width");

    if (args_set("height")) {
      out_height = args_get_positive_integer("height");

      if (args_set("depth")) {
        out_depth = args_get_positive_integer("depth");
        k = ((double)out_width / in_width + (double)out_height / in_height
            + (double)out_depth / in_depth) / 3;
      } else {
        k = ((double)out_width / in_width + (double)out_height / in_height) / 2;
        out_depth = (ulong)ceil(k * in_depth);
      }
    } else {
      if (args_set("depth")) {
        out_depth = args_get_positive_integer("depth");
        k = ((double)out_width / in_width + (double)out_depth / in_depth) / 2;
      } else {
        k = (double)out_width / in_width;
        out_depth = (ulong)ceil(k * in_depth);
      }

      out_height = (ulong)ceil(k * in_height);
    }
  } else if (args_set("height")) {
    out_height = args_get_positive_integer("height");

    if (args_set("depth")) {
      out_depth = args_get_positive_integer("depth");
      k = ((double)out_height / in_height + (double)out_depth / in_depth) / 2;
    } else {
      k = (double)out_height / in_height;
      out_depth = (ulong)ceil(k * in_depth);
    }

    out_width = (ulong)ceil(k * in_width);
  } else if (args_set("depth")) {
    out_depth = args_get_positive_integer("depth");
    k = (double)out_depth / in_depth;
    out_width = (ulong)ceil(k * in_width);
    out_height = (ulong)ceil(k * in_height);
  } else {
    throw("Must specify at least one of --scale-factor, "
        "--recip-scale-factor, --width, or --height");
    k = 0;
    out_width = 0;
    out_height = 0;
    out_depth = 0;
  }

  extra_sharp = args_get_bounded_real("extra-sharpening", -1, DBL_MAX);

  if (k >= 1 && extra_sharp != 0) {
    slog(_l, "Sharpening");
    double_voxmap_new_sharpen(
        &_dtemp,
        _din,
        extra_sharp);
    double_voxmap_delete(&_din);
    donate(&_dtemp, &_din);
    flog(_l);
  }

  slog(_l, "Resizing");
  double_voxmap_new_resize(
      &_dout,
      _din,
      out_width,
      out_height,
      out_depth,
      k,
      args_get_string("method"),
      (byte)args_get_bounded_integer("a", 1, 6),
      (byte)args_get_bounded_integer("v", 1, 7));
  double_voxmap_delete(&_din);
  flog(_l);

  if (k < 1 && extra_sharp != 0) {
    slog(_l, "Sharpening");
    double_voxmap_new_sharpen(
        &_dtemp,
        _dout,
        extra_sharp);
    double_voxmap_delete(&_dout);
    donate(&_dtemp, &_dout);
    flog(_l);
  }

  slog(_l, "Converting double voxmap to byte");
  byte_voxmap_new_from_double(&_out, _dout);
  double_voxmap_delete(&_dout);
  flog(_l);

  stub = args_get_string("output-filename-stub");
  slog(_l, "Saving %s.pvl.nc", stub);
  byte_voxmap_save_pvl_nc(_out, stub);
  flog(_l);

  byte_voxmap_delete(&_out);

  logger_delete(&_l);

  return 0;
}
